/**
 * @author N. Kornweibel
 * @version 1.0
 * 
 * A JPanel for controlling a DFSFileTree.
 * 
 * TODO: refactoring: remove depedency on GasganoDisplay.
 * 
 * This class implements the FileImageDisplayHandler permitting it
 * to act as a Gasgano GUI Service. This is done through a bridge to 
 * external viewers.
 * 
 */

package org.eso.gasgano;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JToolBar;
import javax.swing.SwingUtilities;
import javax.swing.ToolTipManager;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.LineBorder;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.TreePath;

import org.eso.dfs.gui.RecallingComboBox;
import org.eso.dfs.services.ServiceException;
import org.eso.dfs.services.gui.FileImageDisplayHandler;
import org.eso.gasgano.datamodel.gui.OBComponent;
import org.eso.gasgano.dfspipeline.DFSPipelineEnvironment;
import org.eso.gasgano.dfspipeline.DFSPipelineStatusListener;
import org.eso.gasgano.gui.DFSFileTreeTable;
import org.eso.gasgano.gui.DFSFileTreeTableModel;
import org.eso.gasgano.gui.GUIUtils;
import org.eso.gasgano.gui.JTreeTable;
import org.eso.gasgano.gui.TreeModelIterator;
import org.eso.gasgano.gui.TreeTableModel;
import org.eso.gasgano.properties.PropertyDB;
import org.eso.gasgano.viewers.ExternalViewer;
import org.eso.gasgano.viewers.ExternalViewerException;
import org.eso.gasgano.viewers.ExternalViewerFactory;

/**
 * @author Nick Kornweibel
 * 
 * Panel for controlling the directory tree view.
 */
public class TreeControlPanel
    extends JPanel
    implements ActionListener, TreeSelectionListener, 
    DFSPipelineStatusListener,
    FileImageDisplayHandler {

    private JComboBox groupSelector;
    private JButton expandTree;
    private RecallingComboBox findEntry;
    
    private JButton findButton;

    private JTreeTable treeTable;
    private TreeTableModel model;

    private static final String groupByProgLabel = "Default grouping";
    private static final String groupByDirLabel = "Group by Directory";
    private static final String groupByTelLabel = "Group by Telescope";
    private static final String expandLabel = "expand";
    private static final String collapseLabel = "collapse";
    private static final String findLabel = "find";

    private JLabel rbsLabel;
    private JLabel doLabel;
    private JLabel selectedNumber;

    private JToolBar toolBar;
    private JButton displayButton;
    private JButton reportButton;
    private JButton scriptBoardButton;
    private GasganoDisplay mainDisplay;

    public TreeControlPanel(
        JTreeTable t,
        TreeTableModel m,
        GasganoDisplay gd) {
        treeTable = t;
        model = m;
        mainDisplay = gd;
        init();
    }

    private void init() {
        Vector items = new Vector(2);
        setLayout(new BorderLayout());

        items.addElement(groupByProgLabel);
        items.addElement(groupByDirLabel);
        items.addElement(groupByTelLabel);
        groupSelector = new JComboBox(items);
        groupSelector.addActionListener(this);

        expandTree = new JButton(expandLabel);
        expandTree.addActionListener(this);

        JPanel searchPanel = new JPanel();

        searchPanel.add(new JLabel("Find entry:"));
        findEntry = new RecallingComboBox(true, 20);
        findEntry.addActionListener(this);
        searchPanel.add(findEntry);
        findButton = new JButton(findLabel);
        findButton.addActionListener(this);
        searchPanel.add(findButton);
        searchPanel.setBorder(new LineBorder(Color.black, 1));
        JPanel centerPanel = new JPanel();

        centerPanel.add(groupSelector);
        centerPanel.add(expandTree);
        centerPanel.add(searchPanel);

        add(centerPanel, "Center");

        JPanel eastPanel = new JPanel(new BorderLayout());
        selectedNumber = new JLabel("0");
        selectedNumber.setBorder(new EmptyBorder(3, 1, 10, 6));
        treeTable.getTree().addTreeSelectionListener(this);
        eastPanel.add(selectedNumber, "West");

        if (DFSPipelineEnvironment.getInstance().isPipelineEnvironment()) {
            Font tmpFont = null;
            doLabel = new JLabel("DO");
            doLabel.setHorizontalTextPosition(JButton.LEFT);

            tmpFont =
                new Font(
                    doLabel.getFont().getName(),
                    Font.BOLD,
                    doLabel.getFont().getSize());

            // doButton.setIcon( new ImageIcon(TreeControlPanel.class.getResource("exit.jpg")));
            // doButton.setEditable(false);
            rbsLabel = new JLabel("RBS");
            rbsLabel.setHorizontalTextPosition(JButton.LEFT);

            doLabel.setFont(tmpFont);
            rbsLabel.setFont(tmpFont);
            // rbsButton.setIcon( new ImageIcon(TreeControlPanel.class.getResource("flag.jpg")));

            // doButton.setEditable(false);

            JPanel infrastructurePanel = new JPanel();
            infrastructurePanel.setBorder(
                new CompoundBorder(
                    new EmptyBorder(2, 2, 2, 5),
                    new CompoundBorder(
                        new LineBorder(Color.blue),
                        new EmptyBorder(1, 5, 1, 5))));
            infrastructurePanel.setLayout(new GridLayout(2, 1));
            infrastructurePanel.add(rbsLabel);
            infrastructurePanel.add(doLabel);

            eastPanel.add(infrastructurePanel, "East");
        }

        toolBar = new JToolBar();
        toolBar.setFloatable(false);
        toolBar.setBorder(new EmptyBorder(1, 10, 1, 1));
        ToolTipManager.sharedInstance().setInitialDelay(50);

        ImageIcon tmpIcon = new ImageIcon(getClass().getResource(GUIUtils.videoGif));
        displayButton = new JButton(tmpIcon);

        tmpIcon = new ImageIcon(getClass().getResource(GUIUtils.documentGif));
        reportButton = new JButton(tmpIcon);
        tmpIcon = new ImageIcon(getClass().getResource(GUIUtils.disknavGif));
        scriptBoardButton = new JButton(tmpIcon);

        displayButton.addActionListener(this);
        scriptBoardButton.addActionListener(this);
        reportButton.addActionListener(this);

        displayButton.setToolTipText("External Viewer");
        scriptBoardButton.setToolTipText("Script Board");
        reportButton.setToolTipText("Frames Report");

        toolBar.add(displayButton);
        toolBar.add(reportButton);
        toolBar.add(scriptBoardButton);

        add(toolBar, "West");
        add(eastPanel, "East");
    }

    /**
     * event callback for DFSPipelineStatusListener interface.
     * sets some visible signal for the do buttons.
     * @see org.eso.gasgano.dfspipeline.DFSPipelineStatusListener#notifyDORunning(boolean)
     */
    public void notifyDORunning(boolean status) {
        Color activeColor = Color.blue;
        if (status) {
            doLabel.setForeground(activeColor);
        } else {
            doLabel.setForeground(Color.white);
        }
    }
    
    /**
     * event callback for DFSPipelineStatusListener interface.
     * sets some visible signal for the rbs buttons.
     * @see org.eso.gasgano.dfspipeline.DFSPipelineStatusListener#notifyRBSRunning(boolean)
     */
    public void notifyRBSRunning(boolean status) {
        Color activeColor = Color.blue;
        if (status) {
            rbsLabel.setForeground(activeColor);
        } else {
            rbsLabel.setForeground(Color.white);
        }
    }

    /**
     * ActionListener interface implementation - this should be removed
     * and converted to use the Action class model.
     * @see java.awt.event.ActionListener#actionPerformed(ActionEvent)
     */
    public void actionPerformed(ActionEvent a) {
        if (a.getSource() == groupSelector) {
            String item = groupSelector.getSelectedItem().toString();
            if (item == groupByProgLabel) {
                groupBy(0);
                expandTree.setText(expandLabel);
            }
            if (item == groupByDirLabel) {
                groupBy(1);
                expandTree.setText(expandLabel);
            }
            if (item == groupByTelLabel) {
                groupBy(2);
                expandTree.setText(expandLabel);
            }
        } else if (a.getActionCommand() == expandLabel) {
            expandTree.setText(collapseLabel);
            expandTree();

        } else if (a.getActionCommand() == collapseLabel) {
            expandTree.setText(expandLabel);
            collapseTree();
        } else if (
            a.getSource() == findEntry || a.getActionCommand() == findLabel) {
            // Due to multiple events being fired for editable comboboxes, the following
            // check is made:
            // if the selected index is -1, the event does not indicate the end
            // of an item selection....so do nothing until a valid item is selected.
            if (findEntry.getSelectedIndex() > -1) {
                ImageIcon showPic = null;
                if (findEntry.getText().equalsIgnoreCase("gasgano")) {
                    URL loc = Object.class.getResource(GUIUtils.gasganoGif);
                    if (loc != null) {
                        showPic = new ImageIcon(loc);
                    }
                } else if (findEntry.getText().equalsIgnoreCase("podrace")) {
                    URL loc = Object.class.getResource(GUIUtils.podraceJpg);
                    if (loc != null) {
                        showPic = new ImageIcon(loc);
                    }
                }
                if (showPic != null && getTopLevelAncestor() instanceof JFrame) {
                    JFrame parent = (JFrame)getTopLevelAncestor();
                    JOptionPane pane =
                        new JOptionPane(
                            "",
                            JOptionPane.INFORMATION_MESSAGE,
                            JOptionPane.DEFAULT_OPTION,
                            showPic);
                    JDialog dialog = pane.createDialog(parent, "Gasgano");
                    dialog.show();
                }
                if (!searchTree(findEntry.getText())) {
                    userMessage("\'" + findEntry.getText() + "\' not located.");
                }
            }
        } else if (a.getSource() == displayButton) {
            callExternalViewer();
        } else if (a.getSource() == scriptBoardButton) {
            mainDisplay.toggleScriptBoard();
        } else if (a.getSource() == reportButton) {
            mainDisplay.generateReport();
        }
    }
    
    
    /**
     * Search for an OBComponent in the tree with an id matching that
     * given.
     * If found, make the matching node visible to the user.
     * @param obcId - id of the OBComponent to be located.
     * @return boolean
     */
    public boolean searchTreeForOBComponent(String obcId) {
        Object endSearchNode = null;
        boolean rc = false;
        TreePath selPath = treeTable.getTree().getLeadSelectionPath();
        boolean found = false;
        TreeModelIterator iterator = null;
        Object currentNode;
        int columnCount = model.getColumnCount();

        if (selPath != null) {
            endSearchNode = selPath.getLastPathComponent();
        } else {
            endSearchNode = model.getRoot();
        }
        iterator = new TreeModelIterator(model, selPath);

        do {
            iterator.iteratePath();
            currentNode = iterator.getCurrentPath().lastElement();
            if (currentNode instanceof OBComponent) {
                found = ((OBComponent)currentNode).getId().equals(obcId);
            }

        } while (currentNode != endSearchNode && !found);
        if (found) {
            Vector pathVector = iterator.getCurrentPath();
            Object[] path = new Object[pathVector.size()];
            int i = 0;
            while (i < path.length) {
                path[i] = pathVector.elementAt(i);
                i++;
            }
            TreePath tp = new TreePath(path);
            //
            // Nasty hack to fix problem noticed after porting to Java 1.4:
            // If a row that is to be selected (highlighted) must first
            // be expanded, the selection is cleared due to the interaction
            // or data change events between the Tree and Table parts of
            // the JTreeTable.
            // By call setSelectionPath() twice the second call works if
            // the first call did not result in a selected row.
            //
            treeTable.getTree().setSelectionPath(tp);
            treeTable.getTree().setSelectionPath(tp);

            // Following does not work for JTreeTable.
            // treeTable.getTree().scrollPathToVisible(tp);
            // do as per a normal JTable instead:
            Rectangle r =
                treeTable.getCellRect(treeTable.getSelectedRow(), 0, true);
            treeTable.scrollRectToVisible(r);
        }
        return found;
    }

    /**
     * Search for the specified string in the tree structure. If found,
     * make the matching region visible to the user.
     * @param findString
     * @return boolean
     */
    public boolean searchTree(String findString) {
        Object endSearchNode = null;
        boolean rc = false;
        TreePath selPath = treeTable.getTree().getLeadSelectionPath();
        boolean found = false;
        TreeModelIterator iterator = null;
        Object currentNode;
        int columnCount = model.getColumnCount();

        findString = findString.toUpperCase();
        if (selPath != null) {
            endSearchNode = selPath.getLastPathComponent();
        } else {
            endSearchNode = model.getRoot();
        }
        iterator = new TreeModelIterator(model, selPath);

        do {
            iterator.iteratePath();
            currentNode = iterator.getCurrentPath().lastElement();
            for (int i = 0; i < columnCount && !found; i++) {
                Object cellValue = model.getValueAt(currentNode, i);
                if (cellValue != null) {
                    String nodeVal = cellValue.toString().toUpperCase();
                    if (nodeVal.indexOf(findString) != -1) {
                        found = true;
                    }
                }
            }

        } while (currentNode != endSearchNode && !found);
        if (found) {
            Vector pathVector = iterator.getCurrentPath();
            Object[] path = new Object[pathVector.size()];
            int i = 0;
            while (i < path.length) {
                path[i] = pathVector.elementAt(i);
                i++;
            }
            TreePath tp = new TreePath(path);
            //
            // Nasty hack to fix problem noticed after porting to Java 1.4:
            // If a row that is to be selected (highlighted) must first
            // be expanded, the selection is cleared due to the interaction
            // or data change events between the Tree and Table parts of
            // the JTreeTable.
            // By call setSelectionPath() twice the second call works if
            // the first call did not result in a selected row.
            //
            treeTable.getTree().setSelectionPath(tp);
            treeTable.getTree().setSelectionPath(tp);

            // Following does not work for JTreeTable.
            // treeTable.getTree().scrollPathToVisible(tp);
            // do as per a normal JTable instead:
            Rectangle r =
                treeTable.getCellRect(treeTable.getSelectedRow(), 0, true);
            treeTable.scrollRectToVisible(r);
        }
        return found;
    }

    /**
     * Utility method to display a user message to the user.
     * @param msg
     */
    private void userMessage(String msg) {
        GUIUtils.userMessage(this, msg);
    }

    /**
     * Modify the grouping of the mode, if the model is not the filesystem
     * model, nothing happens.
     * @param gbLevel
     */
    public void groupBy(int gbLevel) {
        if (model instanceof DFSFileTreeTableModel) {
            treeTable.getTree().clearSelection();
            mainDisplay.showBlankDetail(true);
            ((DFSFileTreeTableModel) model).setGroupDataByLevel(gbLevel);
        }
    }

    /**
     * Collapse the tree, except for row 0, which is always to be displayed.
     */
    public void collapseTree() {
        // do not collapse row number 0 - we want to show
        // the program-ids.
        int rows = treeTable.getRowCount() - 1;
        for (int i = rows; i > 0; i--) {
            treeTable.getTree().collapseRow(i);
        }
    }

    /**
     * Expand the tree completely.
     */
    public void expandTree() {
        int i = 0;
        for (; i < treeTable.getTree().getRowCount(); i++) {
            treeTable.getTree().expandRow(i);
        }
    }

    public Component getFindEntryComponent() {
        return findEntry;
    }

    /**
     * Implementation method for the TreeSelectionListener()
     */
    public void valueChanged(TreeSelectionEvent e) {
        Vector files = null;
        if (treeTable instanceof DFSFileTreeTable) {
            files = ((DFSFileTreeTable) treeTable).getSelectedOBComponents();
        }
        if (treeTable instanceof ResultsTreeTable) {
            files = ((ResultsTreeTable) treeTable).getSelectedOBComponents();
        }

        if (files == null) {
            selectedNumber.setText("0");
        } else {
            selectedNumber.setText(String.valueOf(files.size()));
        }
    }

    /**
     * send the files selected in the tree to the external viewer.
     * there are three main outcomes:
     * 1. no selected files - error to user
     * 2. selected files, if all fits, are sent to the cube viewer
     * 3. single file, depending on extension, is sent to appropriate display.
     * @return false if files could not be displayed, else true.
     */
    public boolean callExternalViewer() {
        DFSFileTreeTable tree = (DFSFileTreeTable) treeTable;
        List obcomp = tree.getSelectedOBComponents();
        List files = new ArrayList();
        Iterator i = obcomp.iterator();
        while (i.hasNext()) {
            OBComponent obc = (OBComponent)i.next();
            files.add(obc.getId());
        }         
        return callExternalViewer(files);
    }
    
    /**
     * Display the set of specified files names.
     * @param files List of String object paths to files.
     * @return boolean true on success, else false.
     */
    public boolean callExternalViewer(List files) {

        // the parameters needed for the viewer.
        String displayCommand = null;
        String displayCommandType = null;
        Vector displayArguments = new Vector();

        if (files.size() == 0) {
            userMessage("No files have been selected.");
            return false;
        }
        if (files.size() > 1) {
            // it could be a cube
            boolean valid = true;
            for (int i = 0; i < files.size(); i++) {
                String ffoc = (String) files.get(i);
                if (!ffoc.endsWith(".fits")) {
                    valid = false;
                    break;
                } else {
                    // TODO: why is the space needed after the file path?
                    displayArguments.addElement(ffoc+" ");
                }
            }
            if (valid) {
                displayCommand = PropertyDB.getInstance().getProperty("CUBE_TO_VIEWER");
                displayCommandType =
                PropertyDB.getInstance().getProperty("CUBE_VIEWER_TYPE");
                if (displayCommand.equals("")
                    || displayCommandType.equals("")) {
                    userMessage("The viewer for cubes\nhas not been specified");
                    return false;
                }
            } else {
                userMessage("Sorry - I wouldn't know how to display that!\nFor a cube to be displayed all files must be FITS files");
                return false;
            }

        } else { // it's one file only 
            String ffoc = (String)files.get(0);
            if (ffoc.endsWith(".fits")) {
                displayCommand = PropertyDB.getInstance().getProperty("FITS_TO_VIEWER");
                displayCommandType =
                PropertyDB.getInstance().getProperty("FITS_VIEWER_TYPE");

                if (displayCommand.equals("")
                    || displayCommandType.equals("")) {
                    userMessage("The viewer for fits files\nhas not been specified");
                    return false;
                }
                displayArguments.addElement(ffoc);
            }
            if (ffoc.endsWith(".tfits")) {
                displayCommand = PropertyDB.getInstance().getProperty("TFITS_TO_VIEWER");
                displayCommandType =
                PropertyDB.getInstance().getProperty("TFITS_VIEWER_TYPE");
                if (displayCommand.equals("")
                    || displayCommandType.equals("")) {
                    userMessage("The viewer for binary tables\nhas not been specified");
                    return false;
                }
                displayArguments.addElement(ffoc);
            }
        }
        boolean rc = false;
        if (displayCommand == null
            || displayCommandType == null
            || displayArguments.size() == 0) {
            userMessage("Sorry - I wouldn't know\nhow to display that!");
        } else {
            ExternalViewer viewer = null;
            try {
                viewer =
                    ExternalViewerFactory.getInstance().getExternalViewer(
                        displayCommandType);
                viewer.setCommand(displayCommand);
                viewer.displayFiles(displayArguments);
                rc = true;
            } catch (ExternalViewerException e) {
                userMessage("Error with viewer: " + e.getMessage());
            }
        }
        return rc;
    }

    /**
     * Implementation of the FileImageDisplayHandler interface.
     * This class provides a display service acting as a bridge to 
     * external viewers.
     * Note: always returns true as error message are handled internally
     * by the function - returning false will result in multiple error
     * dialogs to the user, which is annoying.
     * @see org.eso.gasgano.services.gui.FileImageDisplayHandler#displayFileImages(java.lang.String[])
     */
    public boolean displayFileImages(String[] fileIds, boolean showImage) throws ServiceException {
        boolean rc = false;
        List files = new ArrayList();
        if (fileIds != null && fileIds.length > 0) {
            for (int i = 0; i < fileIds.length; i++) {
                files.add(fileIds[i]);
            }
            // now locate and show the last file in Gasgano, if it exists in the model
            if (fileIds.length > 0) {
                searchTreeForOBComponent(fileIds[fileIds.length-1]);
                if (!showImage) {
                    SwingUtilities.invokeLater(new Runnable() {
                        public void run() {
                            // if we are not sending the file(s) to the external viewer, then
                            // bring the Gasgano Main Window to the font.
                            // NOTE: toFront() does nothing with XWin32 X-windows emulator, unless
                            // the Main Window is currently iconified.
                            Container frame = getTopLevelAncestor();
                            if (frame != null && frame instanceof JFrame) {
                                ((JFrame)frame).setState(JFrame.NORMAL);
                                ((JFrame)frame).toFront();
                            }
                        }});
                }
            }
            rc = true;
            if (showImage) {
                rc = callExternalViewer(files);
            }
        }
        // return rc;
        return true;
    }
}
