package org.eso.gasgano.datamodel.database;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Vector;

import org.eso.gasgano.datamodel.DataModelChangeNotifier;
import org.eso.gasgano.datamodel.DataModelEvent;
import org.eso.gasgano.datamodel.FileDescription;
import org.eso.gasgano.datamodel.OBComponentFilter;
import org.eso.gasgano.datamodel.ProgramDescription;
import org.eso.gasgano.datamodel.gui.DBOBComponent;
import org.eso.gasgano.keyword.Keyword;
import org.eso.gasgano.properties.GasProp;
import org.eso.gasgano.properties.PropertyDB;
import org.eso.gasgano.tools.PasswordSource;
import org.eso.gasgano.tools.SortedHashtable;

/**
 * DBDataModel: a class representing the data model for DB connection<p>
 * taylored for working with the ESO observation database<p>
 * a Database connection is currently kept alive for all the duration of the
 * program, but there is not fail-safe mechanism yet<p>
 *
 * the building blocks are DBContainers, which act
 * both as programs and observations<p>
 * Currently not provision is made for maintaining more than
 * the Standard Prog/Obs tree View<p>
 * A further Telescope View might be added later<p>
 * 
 * What will not be added is an Event Messaging System,
 * based on DataModelEvent
 * as for the DFSFile Data Model<p>
 *
 * @author: Michele Zamparelli
 * @see: DataModel, DBContainer, DBOBComponent
 * 
 * TODO: refacotr: dependence on GUI layer here (datamode.gui) that
 * should be removed.
 * 
 **/

public class DBDataModel extends DataModelChangeNotifier {
    // set or programs, containing obs, containing components.
    private DBContainer programs = null;
    private DBContainer telescopes = null;
    private ProgramDescription progDescFactory = null;
    private OBComponentFilter fileFilter = null;
    private boolean visualDebug = false;
    private DBTranslation translationTable = null;

    private int filesCount = 0;
    private int initialChunkSize = 500;
    private String errorMessage = null;
    private PasswordSource passwordSource = null;
    private boolean dataModelUpdating = false;
    //private SlidingArray rowsStored = null ; 
    private int knownMaxRow;
    private boolean dataEndReached;

    private String serverURL = null;
    private String userName = null;
    private String passWord = null;
    private String permanentPassWord = null;
    //private String tableName = null ; 
    private Properties connectionProperties = null;
    private Connection permanentConnection = null;

    private DBContainer dummyBox = new DBContainer('a', "Loading");
    private boolean extendedDatabase = false;
    private SortedHashtable instToTable = null;
    private SortedHashtable tableToMap = null;

    public DBDataModel() {
        programs = new DBContainer();
        telescopes = new DBContainer();
        translationTable = new DBTranslation();
        connectionProperties = new Properties();
    }
    
    /**
     * Utility method to determine if debug is on or not.
     * @return boolean
     */
    private static boolean isDebug(int lvl) {
        return PropertyDB.getInstance().isDebug(lvl);
    }

    /**
     * spawns a JDialogBox to ask for a password,<p>
     * load the JDBC Class, establishes the connection<p>
     * 
     * carries out a first search of the database with<p>
     * buildup up the programs list<p>
     *
     * Returns false if any of these steps fails<p>
     *  Notice that this method will be called anytime
     * the user changed the instrument or the period<p>
     */

    public boolean init() {
        programs.clear();
        telescopes.clear();

        instToTable = new SortedHashtable();
        tableToMap = new SortedHashtable();
        serverURL = (String) PropertyDB.getInstance().getProperty("DBOBS_SERVER_URL");
        userName = (String) PropertyDB.getInstance().getProperty("DBOBS_USERID");
        if (permanentPassWord == null) {
            passWord =
                passwordSource.popupPasswordPrompt(
                    userName,
                    "Enter Observations Database password",
                    "DBOBS_REQUIRED");
        }
        //tableName = (String)Application.getProperty("DBOBS_TABLE");
        connectionProperties.put("user", userName);
        connectionProperties.put("password", passWord);

        if (passWord.equals("")) {
            return false;
        }

        try {
            Class.forName("com.sybase.jdbc.SybDriver");
        } catch (ClassNotFoundException cnfe) {
            System.err.println("Could not load Sybase JDBC\n" + cnfe);
            return false;
        }

        if (isDebug(PropertyDB.debugDatabase)) {
            System.out.println("Establishing connection......");
        }

        try {
            permanentConnection =
                DriverManager.getConnection(serverURL, connectionProperties);
        } catch (SQLException sqle) {
            System.err.println("Connection Failed" + sqle);
            return false;
        }

        if (isDebug(PropertyDB.debugDatabase)) {
            System.out.println("Connection established");
        }

        determineExtensions();

        dataModelUpdating = true; // set the semaphore
        // rowsStored = new SlidingArray(initialChunkSize );
        // readProgramsFromDatabase(0, initialChunkSize );

        if (createProgramsForDefault() == false) {
            System.err.println(getErrorMessage());
            dataModelUpdating = false;
            return false;
        }

        if (isDebug(PropertyDB.debugDatabase)) {
            System.err.println("DBDataModel.init completed");
        }

        dataModelUpdating = false;
        permanentPassWord = passWord;
        return true;
    }

    public void cleanup() {
        if (programs != null) {
            programs.clear();
        }
        if (telescopes != null) {
            telescopes.clear();
        }
        if (instToTable != null) {
            instToTable.clear();
        }
        if (tableToMap != null) {
            tableToMap.clear();
        }
        permanentPassWord = null;
        if (permanentConnection != null) {
            try {
                permanentConnection.close();
            } catch (SQLException sqle) {
                System.err.println("Could not close connection: " + sqle);
            }
        }
    }

    /**
     * tentative for a semaphor
     * @returns: a semaphor to indicate that the
     * update method is still running
     *
     */

    public boolean isUpdating() {
        return dataModelUpdating;
    }

    public int getFilesCount() {
        return filesCount;
    }

    public OBComponentFilter getFileFilter() {
        return fileFilter;
    }

    public void setPasswordSource(PasswordSource ps) {
        passwordSource = ps;
    }

    /**
     * @param p The FileDescription object to be used to configure the
     * descriptions of OBComponent objects managed by this DataModel.
     */
    public void setDefaultFileDescription(FileDescription fd) {
        DBOBComponent.setDefaultFileDescription(fd);
    }

    /**
     * @return The FileDescription object to be used to configure the
     * descriptions of OBComponent objects managed by this DataModel.
     */
    public FileDescription getDefaultFileDescription() {
        return DBOBComponent.getDefaultFileDescription();
    }

    //////////////// PROGRAM ///////////////////////////////////

    public DBContainer findProgram(String id) {
        return (DBContainer) programs.get(id);
    }

    public DBContainer findProgram(DBContainer progSet, String id) {
        return (DBContainer) progSet.get(id);
    }

    public int getProgramCount() {
        return programs.size();
    }

    public DBContainer getProgramAt(int i) {
        return (DBContainer) (programs.elementAt(i));
    }

    public int getIndexOfProgram(DBContainer prog) {
        return programs.indexOf(prog);
    }

    public DBContainer getProgramSet() {
        return programs;
    }

    public DBContainer addProgram(String id) {
        return addProgram(programs, id);
    }

    private DBContainer addProgram(DBContainer progSet, String id) {
        DBContainer rc = findProgram(progSet, id);
        if (rc == null) {
            rc = new DBContainer('p', id);

            progSet.put(id, rc);
            rc.setFather(progSet);

            if (isDebug(PropertyDB.debugDataModelChange)) {
                System.out.println(
                    "Added Prog: "
                        + id
                        + " to "
                        + ((progSet == programs)
                            ? "programs"
                            : progSet.toString()));
            }
        }
        return rc;
    }

    /**
     * This procedure populates the model with all the programs<p>
     * present in the database<p>
     *
     * Future extensions might restrict the query to a specific period<p>
     **/

    public boolean createProgramsForDefault() {
        boolean rc = true;
        boolean allPeriods = false;
        boolean allInstruments = false;
        Statement stmt = null;
        char[] programPrefix = { ' ', '1', '2' };
        int loopEnd = 3;
        String periodQuery = null;
        String instQuery = "";

        StringBuffer queryThree =
            new StringBuffer("select distinct prog_id, tel_id from observations..data_products ");

        String startPeriod = PropertyDB.getInstance().getProperty("DBOBS_PERIOD");
        if (startPeriod.equals("All")) {
            allPeriods = true;
            loopEnd = 1;
        }

        String instSelection = PropertyDB.getInstance().getProperty("DBOBS_INSTRUMENT");
        if (instSelection.equals("All")) {
            allInstruments = true;
        }

        if (allInstruments == false) {
            if (allPeriods) {
                instQuery = "where ins_id = \'" + instSelection + "\'";
            } else {
                instQuery = "and  ins_id = \'" + instSelection + "\'";
            }
        }

        // loop over the three classes of programs for each period
        for (int z = 0; z < loopEnd; z++) {
            if (z == 0) {
                periodQuery = "where prog_id like \'" + startPeriod + "%\'";
                if (allPeriods) {
                    periodQuery = "";
                }
            } else {
                periodQuery =
                    "where prog_id like \'"
                        + programPrefix[z]
                        + startPeriod
                        + "%\'";
            }

            try {
                stmt = permanentConnection.createStatement();
                ResultSet rs = null;

                if (isDebug(PropertyDB.debugDatabase)) {
                    System.out.println(
                        queryThree.toString() + periodQuery + instQuery);
                }

                boolean results =
                    stmt.execute(
                        queryThree.toString() + periodQuery + instQuery);
                do {
                    if (results) {
                        rs = stmt.getResultSet();
                        for (int rowNum = 1; rs.next(); rowNum++) {
                            String progIdString = rs.getString(1);
                            String telString = rs.getString(2);
                            if (progIdString != null) {
                                DBContainer prog = addProgram(progIdString);

                                if (isDebug(PropertyDB.debugDatabase)) {
                                    System.out.println(progIdString);
                                }

                                if (telString != null) {
                                    DBContainer telescope =
                                        addTelescope(telString);
                                    addProgram(telescope, progIdString);
                                }
                            }
                        }
                    }
                    results = stmt.getMoreResults();
                } while (results);
                stmt.close();
            } catch (SQLException sqe) {
                errorMessage =
                    new String(
                        sqe.toString() + ", sqlstate = " + sqe.getSQLState());
                rc = false;
                // stmt.cancel();
                connectionLost(sqe);
            }
        }
        return rc;
    }

    /**
     * 
     *
     */

    public void addSingleProgram(String program) {
        boolean debug = false;
        Statement stmt = null;
        DataModelEvent modelChangedEvent = null;
        StringBuffer queryOne =
            new StringBuffer("select distinct prog_id, tel_id from observations..data_products where prog_id = \'");

        if (isDebug(PropertyDB.debugDatabase)) {
            debug = true;
        }

        try {
            stmt = permanentConnection.createStatement();
            ResultSet rs = null;
            if (debug) {

            }

            boolean results =
                stmt.execute(queryOne.toString() + program + "\'");
            do {
                if (results) {
                    rs = stmt.getResultSet();
                    for (int rowNum = 1; rs.next(); rowNum++) {
                        String progIdString = rs.getString(1);
                        String telString = rs.getString(2);
                        if (progIdString != null) {
                            DBContainer prog = addProgram(progIdString);

                            if (debug) {
                                System.out.println(progIdString);
                            }
                            if (telString != null) {
                                DBContainer telescope = addTelescope(telString);
                                addProgram(telescope, progIdString);
                            }
                            modelChangedEvent =
                                new DataModelEvent(
                                    DataModelEvent.PROGRAM_ADDED,
                                    prog);
                            fireDataModelChangeEvent(modelChangedEvent);
                        }
                    }
                }
                results = stmt.getMoreResults();
            } while (results);
            stmt.close();
        } catch (SQLException sqe) {
            errorMessage =
                new String(
                    sqe.toString() + ", sqlstate = " + sqe.getSQLState());
            connectionLost(sqe);
        }
    }

    //////////////// OBSERVATION ///////////////////////////////////

    public DBContainer findObservation(DBContainer progSet, String id) {
        return (DBContainer) progSet.get(id);
    }

    private DBContainer addObservation(
        DBContainer progSet,
        String id,
        String name) {
        DBContainer rc = findObservation(progSet, id);
        if (rc == null) {
            rc = new DBContainer('o', id);
            rc.setDescription(id + " " + name);
            progSet.put(id, rc);
            rc.setFather(progSet);

            if (isDebug(PropertyDB.debugDataModelChange)) {
                System.out.println(
                    "Added Observation: "
                        + id
                        + " to "
                        + ((progSet == programs)
                            ? "programs"
                            : progSet.toString()));
            }
        }
        return rc;
    }

    /**
     * fills the observations for a given  program
     * counts the amount of frames for each observation
     * within a program
     *
     * @params: the program for which Observations should be created
     * @returns: false on failure 
     */

    public boolean createObservations(DBContainer prog) {
        boolean rc = true;
        boolean results = false;
        boolean results2 = false;
        String tmpQuery = null;
        ResultSet rs = null;
        //new String("select obs_name from observations..data_products "+  "where prog_id = \'");
        String queryTwo =
            new String("select distinct ob_id, obs_name from observations..data_products where prog_id = \'");
        Statement stmt = null;
        Statement stmt2 = null;

        try {
            stmt = permanentConnection.createStatement();
            tmpQuery = queryTwo + prog + "\' ";
            if (isDebug(PropertyDB.debugDatabase)) {
                System.err.println("creating observations for prog " + prog);
                System.err.flush();
            }
            results = stmt.execute(tmpQuery);
            if (isDebug(PropertyDB.debugDatabase)) {
                System.err.println(tmpQuery + ((results) ? " ok" : " empty"));
                System.err.flush();
            }
            do {
                if (results) {
                    rs = stmt.getResultSet();
                    for (int rowNum = 1; rs.next(); rowNum++) {
                        String obIdString = rs.getString(1);
                        String obName = rs.getString(2);
                        if (obIdString != null) {
                            if (isDebug(PropertyDB.debugDatabase)) {
                                System.err.println(
                                    "         " + obIdString + "  " + obName);
                            }
                            DBContainer obs =
                                addObservation(prog, obIdString, obName);
                        }
                    }
                } else {
                    System.out.println("........No observations");
                }
                results = stmt.getMoreResults();
            } while (results);
            stmt.close();
        } catch (SQLException sqe) {
            errorMessage =
                new String(
                    sqe.toString() + ", sqlstate = " + sqe.getSQLState());
            rc = false;
            connectionLost(sqe);
        }
        return rc;
    }

    /**
     * pick up all frame belonging to a certain observation and
     * prog 
     */

    public boolean fillObservation(DBContainer prog, DBContainer ob) {
        boolean rc = false;
        String queryTwo = null;
        DBContainer program, observation = null;
        Statement stmt = null;
        boolean treeTelescope = true;

        queryTwo =
            "select * from observations..data_products "
                + "where prog_id = \'"
                + prog.getId()
                + "\' and ob_id = "
                + ob.getId();

        if (prog == null) {
            System.out.println("Troubles finding DBContainer for " + prog);
        }

        if (isDebug(PropertyDB.debugDatabase)) {
            System.err.println(
                "Finding  frames for Program " + prog + " observation " + ob);
        }

        if (prog.getFather() == programs) {
            treeTelescope = false;
        }

        try {
            stmt = permanentConnection.createStatement();
            ResultSet rs = null;
            if (isDebug(PropertyDB.debugDatabase)) {
                System.err.println(queryTwo);
            }
            boolean results = stmt.execute(queryTwo);

            do {
                if (results) {
                    rs = stmt.getResultSet();
                    for (int rowNum = 1; rs.next(); rowNum++) {
                        DBOBComponent tmpComp = new DBOBComponent();
                        loadPrimaryKeywords(tmpComp, rs);
                        if (extendedDatabase) {
                            addExtendedKeywords(tmpComp);
                        }
                        tmpComp.init();
                        tmpComp.freeHeaderMemory();
                        ob.put(tmpComp.getId(), tmpComp);
                        if (treeTelescope) {
                            tmpComp.setObservation(
                                ob,
                                GasProp.TREE_TELESCOPE);
                        } else {
                            tmpComp.setObservation(
                                ob,
                                GasProp.TREE_DEFAULT);
                        }
                        if (isDebug(PropertyDB.debugDatabase)) {
                            System.err.println(tmpComp);
                        }
                    }
                    results = stmt.getMoreResults();
                }
            } while (results);
            stmt.close();
        } catch (SQLException sqe) {
            errorMessage =
                new String(
                    sqe.toString() + ", sqlstate = " + sqe.getSQLState());
            rc = false;
            connectionLost(sqe);
        }
        return rc;
    }

    public void removeObservations(DBContainer prog) {
        DBContainer tmpCont = null;
        for (Enumeration e = prog.elements(); e.hasMoreElements();) {
            tmpCont = (DBContainer) e.nextElement();
            prog.remove(tmpCont.getId());
        }
    }

    ///////////////////////// TELESCOPES ///////////////////////////////////////////////

    public DBContainer getTelescopesSet() {
        return telescopes;
    }

    public DBContainer getTelescope(String tel) {
        return (DBContainer) telescopes.get(tel);
    }

    public int getTelescopeCount() {
        return telescopes.size();
    }

    public DBContainer getTelescopeAt(int i) {
        return (DBContainer) (telescopes.elementAt(i));
    }

    public int getIndexOfTelescope(DBContainer prog) {
        return telescopes.indexOf(prog);
    }

    public DBContainer findTelescope(String telName) {
        return (DBContainer) telescopes.get(telName);
    }

    public DBContainer addTelescope(String telName) {
        DBContainer rc = findTelescope(telName);
        if (rc == null) {
            rc = new DBContainer('t', telName);
            telescopes.put(telName, rc);
        }
        if (isDebug(PropertyDB.debugDataModelChange)) {
            System.out.println("Added Telescope: " + telName);
        }
        return rc;
    }

    /////////////////////////////////////////////////////////////////////////////////////

    public void expandContainer(DBContainer tmpCont) {
        // this has to be modified if more than one tree view will be introduced
        if (tmpCont.isProgram()) {
            tmpCont.clear();
            createObservations(tmpCont);
        }
        if (tmpCont.isObservation()) {
            //System.out.println("Expanding Observation......");
            if (tmpCont == dummyBox) {
                return;
            }
            fillObservation(tmpCont.getFather(), tmpCont);
        }
        if (tmpCont.isTelescope()) {
            // System.out.println("Expanding Telescope......");
        }
    }

    public void collapseContainer(DBContainer tmpCont) {
        // this has to be modified if more than one tree view will be introduced
        if (tmpCont.isProgram()) {
            tmpCont.clear();
            removeObservations(tmpCont);
            addObservation(tmpCont, dummyBox.getId(), "pippo");
        }

        if (tmpCont.isObservation()) {
            tmpCont.clear();
            removeFrames(tmpCont.toString());
            // slightly misleading, but working
            addObservation(tmpCont, dummyBox.getId(), "pippo");
        }
        // for a telescope nothing is to be done
    }

    public void removeFrames(String observation) {
        // to be filled in ASAP

    }

    /**
     * @return null if there is no error, else and error message.
     */
    public String getErrorMessage() {
        return errorMessage;
    }

    private void connectionLost(SQLException sqle) {
        System.err.println(getErrorMessage());

        boolean lost = true;
        while (lost) {
            try {
                permanentConnection =
                    DriverManager.getConnection(
                        serverURL,
                        connectionProperties);
                lost = false;
            } catch (Exception se) {
                System.err.println("Connection Failed " + se);
            }
            try {
                Thread.sleep(5000);
            } catch (InterruptedException ie) {

            }
            // java.sql.SQLException: JZ0C0: Connection is already closed.
        }
    }

    /**
     * tries to determine in which kind of database we are operating,<p>
     * in particular whether it contains the lookup table 'obs_tables'<p>
     * and if additional tables are available for which instruments<p>
     * This info is stored in the instToTable hashtable<p>
     **/

    private void determineExtensions() {
        Statement stmt = null;
        boolean debug = false;
        String oneQuery =
            "select name from observations..sysobjects where name = \'obs_tables\' and type = \'U\'";
        String twoQuery =
            "select * from observations..obs_tables where instrume = \'";
        String threeQuery = "\' and name != \'data_products\'";
        String fourQuery =
            "select name, keyword from observations..obs_columns where tname = \'";

        if (isDebug(PropertyDB.debugDatabase)) {
            debug = true;
        }

        try {
            stmt = permanentConnection.createStatement();
            ResultSet rs = null;

            if (isDebug(PropertyDB.debugDatabase)) {
                System.out.println(oneQuery);
            }

            boolean results = stmt.execute(oneQuery);
            do {
                if (results) {
                    rs = stmt.getResultSet();
                    for (int rowNum = 1; rs.next(); rowNum++) {
                        String tmpString = rs.getString(1);
                        extendedDatabase = true;
                    }
                }
                results = stmt.getMoreResults();
            } while (results);
            stmt.close();
        } catch (SQLException sqe) {
            errorMessage =
                new String(
                    sqe.toString() + ", sqlstate = " + sqe.getSQLState());
            // stmt.cancel();
            connectionLost(sqe);
        }
        if (!extendedDatabase) {
            return;
        }

        if (debug) {
            System.out.println("This is an extended database");
        }

        for (int i = 0;
            i < GasProp.getProperties().dbInstruments.size();
            i++) {
            String instrume =
                (String) GasProp.getProperties().dbInstruments.elementAt(i);
            String query = twoQuery + instrume + threeQuery;

            if (debug)
                System.out.print(instrume + ": ");

            Vector tmpVector = new Vector();
            try {
                stmt = permanentConnection.createStatement();
                ResultSet rs = null;

                if (debug) {
                    System.out.println(query);
                }
                boolean results = stmt.execute(query);
                do {
                    if (results) {
                        rs = stmt.getResultSet();
                        for (int rowNum = 1; rs.next(); rowNum++) {
                            String tmpString = rs.getString(1);
                            if (debug)
                                System.out.print(tmpString + " ");
                            tmpVector.addElement(
                                new SortedHashtable(tmpString));
                        }
                    }
                    results = stmt.getMoreResults();
                } while (results);
                stmt.close();
            } catch (SQLException sqe) {
                errorMessage =
                    new String(
                        sqe.toString() + ", sqlstate = " + sqe.getSQLState());
                connectionLost(sqe);
            }
            instToTable.put(instrume, tmpVector);
            if (debug)
                System.out.println("");
        }

        for (int i = 0;
            i < GasProp.getProperties().dbInstruments.size();
            i++) {
            String instrume =
                (String) GasProp.getProperties().dbInstruments.elementAt(i);
            Vector tmpVector = (Vector) instToTable.get(instrume);
            for (int j = 0; j < tmpVector.size(); j++) {
                SortedHashtable tmpMap =
                    (SortedHashtable) tmpVector.elementAt(j);
                String query = fourQuery + tmpMap.toString() + "\' ";

                try {
                    stmt = permanentConnection.createStatement();
                    ResultSet rs = null;

                    if (debug) {
                        System.out.println(query);
                    }
                    boolean results = stmt.execute(query);
                    do {
                        if (results) {
                            rs = stmt.getResultSet();
                            for (int rowNum = 1; rs.next(); rowNum++) {
                                String tmpColumn = rs.getString(1);
                                String tmpKeyword = rs.getString(2);
                                if (tmpColumn != null && tmpKeyword != null) {
                                    if (debug)
                                        System.out.println(
                                            tmpColumn
                                                + " =>  "
                                                + Keyword.convertFullNameToDot(
                                                    tmpKeyword));
                                    tmpMap.put(
                                        tmpColumn,
                                        Keyword.convertFullNameToDot(
                                            tmpKeyword));
                                }
                            }
                        }
                        results = stmt.getMoreResults();
                    } while (results);
                    stmt.close();
                } catch (SQLException sqe) {
                    errorMessage =
                        new String(
                            sqe.toString()
                                + ", sqlstate = "
                                + sqe.getSQLState());
                    connectionLost(sqe);
                }
                tableToMap.put(tmpMap.toString(), tmpMap);
            } // end of loop over tables 
        } // end of loop over instruments
    }

    /**
     * this method loads the keywords contained in the all-purpose table
     * 'data_products' into the given DBOBComponent<p>
     *
     * This method makes use of the static table translationTable which is needed
     * in case the database is not an extended one<p>
     * @params: the DBOBComponent to be filled in, the ResultsSet of the running query<p>
     * 
     */

    void loadPrimaryKeywords(DBOBComponent dbComp, ResultSet rs) {
        boolean debug = false;

        if (isDebug(PropertyDB.debugDatabase)) {
            debug = true;
        }

        try {
            for (int z = 0; z < translationTable.size(); z++) {
                String key = translationTable.getKeyAt(z);
                String value = rs.getString((String) translationTable.get(key));
                Keyword keyW = new Keyword(key, value);
                // System.err.println("Adding "+key + " " + value + " " + keyW );
                dbComp.addKeyword(keyW);
            }
        } catch (SQLException sqle) {
            System.err.println(sqle);
        }
    }

    /**
     * all the keywords contained in instrument-specific additional
     * tables are added by this method<p>
     * It make use of the SortedHashtables containing the column-keyword
     * mapping, build in the method determineExtensions<p>
     *
     * @param: the dbComp to have additional keywords set
     */

    void addExtendedKeywords(DBOBComponent dbComp) {
        String instSelection = PropertyDB.getInstance().getProperty("DBOBS_INSTRUMENT");
        Statement stmt = null;
        boolean debug = false;
        boolean allInstruments = false;
        String oneQuery = "select * from observations..";
        String twoQuery = " where dp_id = \'";
        String dp_id = dbComp.getKeyword(Keyword.ARCFILE).getValue();

        if (isDebug(PropertyDB.debugDatabase)) {
            debug = true;
        }

        if (instSelection.equals("All")) {
            instSelection = dbComp.getKeyword(Keyword.INSTRUMENT).getValue();
        }

        Vector tmpVector = (Vector) instToTable.get(instSelection);
        // loop over each table
        for (int j = 0; j < tmpVector.size(); j++) {
            SortedHashtable tmpMap = (SortedHashtable) tmpVector.elementAt(j);
            String query =
                oneQuery + tmpMap.toString() + twoQuery + dp_id + "\'";
            try {
                stmt = permanentConnection.createStatement();
                ResultSet rs = null;

                if (debug) {
                    System.out.println(query);
                }
                boolean results = stmt.execute(query);
                do {
                    if (results) {
                        rs = stmt.getResultSet();
                        for (int rowNum = 1; rs.next(); rowNum++) {
                            for (Enumeration e = tmpMap.keys();
                                e.hasMoreElements();
                                ) {
                                String tmpColumn = (String) e.nextElement();
                                String tmpKey = (String) tmpMap.get(tmpColumn);
                                String tmpValue = rs.getString(tmpColumn);
                                if (tmpValue != null) {
                                    dbComp.addKeyword(
                                        new Keyword(tmpKey, tmpValue));
                                    if (debug) {
                                        System.out.println(
                                            "== " + tmpKey + " " + tmpValue);
                                    }
                                }
                            }
                        }
                    }
                    results = stmt.getMoreResults();
                } while (results);
                stmt.close();
            } catch (SQLException sqe) {
                errorMessage =
                    new String(
                        sqe.toString() + ", sqlstate = " + sqe.getSQLState());
                connectionLost(sqe);
            }
        }
    }

    /**
     * pick up all frame belonging to a certain observation and
     * prog 
     */

    public boolean selectSingleFrame(DBOBComponent dbob) {
        boolean rc = false;
        String queryTwo = null;
        DBContainer program, observation = null;
        Statement stmt = null;
        boolean treeTelescope = true;

        queryTwo =
            "select * from observations..data_products where dp_id = \'"
                + dbob.getId()
                + "\'";

        if (isDebug(PropertyDB.debugDatabase)) {
            System.err.println("selecting single frame" + dbob.getId());
        }

        try {
            stmt = permanentConnection.createStatement();
            ResultSet rs = null;
            if (isDebug(PropertyDB.debugDatabase)) {
                System.err.println(queryTwo);
            }
            boolean results = stmt.execute(queryTwo);

            do {
                if (results) {
                    rs = stmt.getResultSet();
                    for (int rowNum = 1; rs.next(); rowNum++) {
                        loadPrimaryKeywords(dbob, rs);
                        if (extendedDatabase) {
                            addExtendedKeywords(dbob);
                        }
                    }
                    results = stmt.getMoreResults();
                }
            } while (results);
            stmt.close();
        } catch (SQLException sqe) {
            errorMessage =
                new String(
                    sqe.toString() + ", sqlstate = " + sqe.getSQLState());
            rc = false;
            connectionLost(sqe);
        }
        return rc;
    }

}
