/**
 * @author N. Kornweibel
 * @version 1.0
 * A Dialog for selecting a set of Files or directories
 * using a JFileChooser.
 * This is the primary means to add files to the FileSystem data model.
 * The JFileChooser is equiped with drag drop. The drag drop is
 * not useful at this stage as the file chooser and FileListChooser panels
 * are displayed as model dialogs, preventing the main application from receiving
 * dropped files.
 */

package org.eso.gasgano.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ListModel;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.LineBorder;

import org.eso.dfs.gui.SwingFileChooser;

public class FileListChooser extends JPanel implements ActionListener {

    private Vector currentList = null;
    private JButton deleteButton = null;
    private JButton addButton = null;
    private static final String deleteLabel = "Remove Selected Files";
    private static final String addLabel = "Add File";
    private JFileChooser fileChooser = null;
    private JList fileList = null;

    private boolean emergencyBrake = false;
    private int maxRecursiveDirs = 100;
    
    
    /**
     * Test App.
     * @param args
     */
    public static void main(String[] args) {
        JFrame frame = new JFrame("FileListChooser");
        FileListChooser fc = new FileListChooser();
        Vector vec = new Vector();
        vec.addElement(new String("/data/raw"));
        vec.addElement(new String("/data/raw/1999-05-06"));
        vec.addElement(new String("/data/reduced"));
        fc.setFileList(vec);
        if (JOptionPane
            .showConfirmDialog(
                frame,
                fc,
                "Edit directory set",
                JOptionPane.OK_CANCEL_OPTION,
                JOptionPane.PLAIN_MESSAGE)
            == JOptionPane.OK_OPTION) {
            vec = fc.getFileList();
        }
        for (int i = 0; i < vec.size(); i++) {
            System.out.println(vec.elementAt(i).toString());
        }
        System.exit(0);
    }

    /**
     * Build the panel
     * @see java.lang.Object#Object()
     */
    public FileListChooser() {
        currentList = new Vector();
        fileChooser = new SwingFileChooser();
        fileChooser.addActionListener(this);
        fileChooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
        fileChooser.setApproveButtonText(addLabel);
        fileChooser.setMultiSelectionEnabled(true);
        fileChooser.setDragEnabled(true);
        fileList = new JList();
        fileList.setListData(currentList);

        setLayout(new BorderLayout());

        deleteButton = new JButton(deleteLabel);
        deleteButton.addActionListener(this);
        JPanel buttonPanel = new JPanel(new BorderLayout());
        buttonPanel.setBorder(new EmptyBorder(5, 5, 0, 5));
        buttonPanel.add(deleteButton, "East");
        addButton = new JButton(addLabel);
        addButton.addActionListener(this);
        buttonPanel.add(addButton, "West");
        JPanel listPanel = new JPanel(new BorderLayout());
        listPanel.add(new JLabel("Data Directories and Files:"), "North");
        listPanel.add(
            new JScrollPane(
                fileList,
                JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS),
            "Center");
        listPanel.add(buttonPanel, "South");
        listPanel.setBorder(
            new CompoundBorder(
                new LineBorder(Color.black, 1),
                new EmptyBorder(5, 10, 5, 10)));
        listPanel.setPreferredSize(new Dimension(450, 200));
        add(listPanel, "Center");
    }

    /**
     * Note the Vector set here is copied into a ListModel and so the 
     * Vector returned with getFileList() is not one set in this method.
     */
    public void setFileList(Vector files) {
        currentList = files;
        fileList.setListData(currentList);
    }

    public Vector getFileList() {
        ListModel lm = fileList.getModel();
        Vector rc = new Vector(lm.getSize());
        for (int i = 0; i < lm.getSize(); i++) {
            String filePath = lm.getElementAt(i).toString();
            if (filePath != null && filePath.length() > 0) {
                rc.addElement(filePath);
            }
        }
        return rc;
    }

    public void setFileChooser(JFileChooser fc) {
        fileChooser = fc;
    }

    public JFileChooser getFileChooser() {
        // Do not update the directory set as we do not want to change the 				// current file selection.
        return fileChooser;
    }

    public void actionPerformed(ActionEvent a) {
        if (a.getActionCommand() == addLabel) {
            addFile();
        } else if (a.getActionCommand() == deleteLabel) {
            deleteSelectedFiles();
        }
    }

    private void deleteSelectedFiles() {
        // prompt the user and then delete the selected files.
        ListModel lm = fileList.getModel();
        Vector files = new Vector(lm.getSize());
        for (int i = 0; i < lm.getSize(); i++) {
            String filePath = lm.getElementAt(i).toString();
            if (!fileList.isSelectedIndex(i)
                && filePath != null
                && filePath.length() > 0) {
                files.addElement(filePath);
            }
        }
        setFileList(files);
    }

    private void addFile() {
        JFileChooser fChooser = getFileChooser();
        fChooser.setMultiSelectionEnabled(true);
        if (fChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
            File[] selectedFiles = null;
            selectedFiles = fChooser.getSelectedFiles();

            for (int f = 0; f < selectedFiles.length; f++) {
                File[] fileSet;
                if (selectedFiles[f].isDirectory()) {
                    // enter the directory and all sub-dirs.
                    Vector allTheDirs = new Vector();
                    if (getAllTheDirectories(selectedFiles[f], allTheDirs)) {
                        fileSet = new File[allTheDirs.size()];
                        allTheDirs.copyInto(fileSet);
                    } else {
                        // directory could not be expanded to sub-dirs,
                        // so not file is entered (user has already been informed).
                        fileSet = new File[0];
                    }

                } else {
                    // enter the file.
                    fileSet = new File[1];
                    fileSet[0] = selectedFiles[f];
                }

                ListModel lm = null;
                Vector files = getFileList();
                for (int i = 0; i < fileSet.length; i++) {
                    if (fileSet[i] != null) {
                        String selectedFile = null;
                        selectedFile = fileSet[i].getPath();
                        if (files.contains(selectedFile) == false) {
                            files.addElement(selectedFile);
                        }
                    }
                }
                setFileList(files);
            }
        }
    }

    /**
     * Warning: these are recursive procedures!
     *
     *
     */

    private boolean getAllTheDirectories(File father, Vector allTheDirs) {
        allTheDirs.removeAllElements();
        rec_traverse(father, allTheDirs);
        if (emergencyBrake) {
            // notify
            GUIUtils.userMessage(
                this,
                "Too many directories under "
                    + father
                    + ".\nSelection aborted");
            allTheDirs.removeAllElements();
            emergencyBrake = false;
            return false;
        }
        return true;
    }

    private void rec_traverse(File father, Vector allTheDirs) {
        File tmpFile = null;
        String[] childrenFiles = null;
        String[] childrenDirs = null;

        childrenFiles = father.list();
        if (childrenFiles == null || emergencyBrake) {
            return;
        }
        allTheDirs.addElement(father);
        if (allTheDirs.size() > maxRecursiveDirs) {
            emergencyBrake = true;
            return;
        }
        childrenDirs = father.list(new DirsOnlyFilter());
        if (childrenDirs == null) {
            return;
        }
        for (int i = 0; i < childrenDirs.length; i++) {
            try {
                tmpFile =
                    new File(
                        father.getCanonicalPath()
                            + File.separator
                            + childrenDirs[i]);
            } catch (IOException ioe) {
                continue;
            }
            rec_traverse(tmpFile, allTheDirs);
        }
    }

    protected class DirsOnlyFilter implements FilenameFilter {
        public boolean accept(File dir, String name) {
            if (name.startsWith(".")) {
                return false;
            }
            if (new File(dir.toString() + File.separator + name)
                .isDirectory()) {
                return true;
            }
            return false;
        }
    }

}
// end of FileListChooser.java.
