package org.eso.gasgano.gui;

import java.awt.BorderLayout;
import java.awt.Cursor;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.border.TitledBorder;

import org.eso.cpl.Recipe;
import org.eso.cpl.RecipeLibrary;
import org.eso.cpl.gui.ExtensionFileFilter;
import org.eso.cpl.gui.RecipeSelector;
import org.eso.cpl.jni.LTDLException;
import org.eso.cpl.jni.LibraryLoader;
import org.eso.dfs.gui.SwingFileChooser;
import org.eso.gasgano.properties.GasProp;

/**
 * @author Nick Kornweibel
 *
 * Panel for editing the set of CPL libraries used by Gasgano.
 */
public class RecipePropertyPanel extends PropertyEditorSubPanel {

    private GasProp gpi;
    // should be accessed through getLoader(), not directly.
    private LibraryLoader loader;
    private JTable table;
    private ListAdaptor model;

    private boolean edited = false;
    private static JFileChooser chooser;

    /**
     * List of  LibarayDescription objects.
     */
    private List data;

    public RecipePropertyPanel(GasProp gp) {
        gpi = gp;
        data = getRecipeList(gpi);
        model = new ListAdaptor(data, "Recipe Libraries") {
            public boolean isCellEditable(int row, int col) {
                return false;
            }
        };
        table = new JTable(model);
        table.addMouseListener(new MouseAdapter() {
            public void mousePressed(MouseEvent e) {
                if ((e.getClickCount()) == 2) {
                    int row = table.rowAtPoint(new Point(e.getX(), e.getY()));
                    LibraryDescription lib = (LibraryDescription) data.get(row);
                    showRecipeSelector(lib);
                }
            }

        });
        JButton removeButton =
            new JButton(new AbstractAction("Remove Selected") {
            public void actionPerformed(ActionEvent e) {
                edited = true;
                deleteSelectedLibrary(table);
            }
        });

        JButton addButton = new JButton(new AbstractAction("Add Recipe") {
            public void actionPerformed(ActionEvent e) {
                findRecipe();
            }
        });
        JPanel buttons = new JPanel();
        buttons.add(addButton);
        buttons.add(removeButton);
        JScrollPane scroller = new JScrollPane(table);
        scroller.setVerticalScrollBarPolicy(
            JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);

        JPanel tablePanel = new JPanel();

        tablePanel.setBorder(new TitledBorder("Recipe Configuration"));
        tablePanel.setLayout(new BorderLayout());
        tablePanel.add(scroller, BorderLayout.CENTER);
        tablePanel.add(buttons, BorderLayout.SOUTH);
        setLayout(new BorderLayout());
        add(tablePanel, BorderLayout.NORTH);
    }

    /**
     * @see org.eso.gasgano.gui.PropertyEditorSubPanel#loadProperties(org.eso.gasgano.properties.GasProp)
     */
    public void loadProperties(GasProp gp) {
        gpi = gp;
        data = getRecipeList(gpi);
        model.setData(data);
    }

    /**
     * Get a List of LibraryDescription objects from the GasProp given.
     * @param gp
     * @return List
     */
    private static List getRecipeList(GasProp gp) {
        List rc = new ArrayList();
        Iterator i = gp.getRecipeLibraryNames();
        while (i.hasNext()) {
            String lib = (String) i.next();
            List recNames = gp.getRecipesForLibrary(lib);
            // add the entry to the vector to be displayed.
            rc.add(new LibraryDescription(lib, recNames));
        }
        return rc;
    }

    /**
     * Save the settings back to the properties DB.
     * @see org.eso.gasgano.gui.PropertyEditorSubPanel#saveProperties()
     */
    public String saveProperties() {
        String rc = null;
        if (edited) {
            StringBuffer prop = new StringBuffer();
            Iterator i = data.iterator();
            while (i.hasNext()) {
                LibraryDescription d = (LibraryDescription) i.next();
                if (prop.length() != 0) {
                    prop.append(";");
                }
                prop.append(d.toPropertyString());
            }
            gpi.setProperty(GasProp.RECIPE_SET, prop.toString());
        }
        return rc;
    }

    /**
     * @see org.eso.gasgano.gui.PropertyEditorSubPanel#restartRequired()
     */
    public int restartRequired() {
        return edited ? RESTART_ON_SAVE : NOACTION_ON_SAVE;
    }

    /*
     * create and get the library loader
     */
    private LibraryLoader getLoader() throws LTDLException {
        if (loader == null) {
            loader = new LibraryLoader();
        }
        return loader;
    }

    /*
     * release the library loader.
     */
    private void releaseLoader() throws LTDLException {
        if (loader != null) {
            loader.close();
            loader = null;
        }
    }

    /**
     * Add a library of recipes to the set. This is done as follows:
     * - show file chooser to permit user to load linrary
     * - attempt to load library and determine if it is a CPL Recipe
     * - if recipe, display in recipe selector
     * - if user confirms, add linrary and recipes to table.
     */
    public void findRecipe() {
        if (chooser == null) {
            chooser = new SwingFileChooser();
            chooser.setFileFilter(
                new ExtensionFileFilter(
                    getDLLExtension(),
                    "CPL recipe libraries"));
            chooser.setMultiSelectionEnabled(true);
            chooser.setFileHidingEnabled(false);
        }
        if (chooser.showDialog(this, "OK") == JFileChooser.APPROVE_OPTION) {
            showRecipeSelector(chooser.getSelectedFiles(), false);
        }
    }

    private void showRecipeSelector(LibraryDescription lib) {
        File dll = new File(lib.getLibraryName());
        if (!dll.exists()) {
            showError(
                "Library " + lib.getLibraryName() + " could not be found.");
        } else {
            File dlls[] = new File[1];
            dlls[0] = dll;
            showRecipeSelector(dlls, true);
        }
    }

    /**
     * show the recipe selector panel for the dll given (assumes the
     * file exists).
     * @param lib
     */
    private void showRecipeSelector(File dlls[], final boolean readOnly) {

        LibraryLoader l = null;
        final RecipeLibrary[] libs = new RecipeLibrary[dlls.length];
        boolean loadError = false;

        String msg;
        if (dlls.length == 1) {
            msg =
                "If the selected file is not a CPL Recipe Library,\n"
                    + "loading it may crash Gasgano.\n"
                    + "Do you wish to continue?";
        } else {
            msg =
                "If the selected files are not CPL Recipe Libraries,\n"
                    + "loading them may crash Gasgano.\n"
                    + "Do you wish to continue?";
        }

        if (JOptionPane
            .showConfirmDialog(
                this,
                msg,
                "Warning: Library Load",
                JOptionPane.OK_CANCEL_OPTION)
            != JOptionPane.OK_OPTION) {
            // user did not want to continue.
            return;
        }
        File dll = null;
        try {
            l = getLoader();

            for (int i = 0; i < dlls.length; i++) {
                dll = dlls[i];
                // Printing to stdout is useful during debugging in case one
                // in the set of dlls fails to load
                System.out.print("Loading "+dll.getName()+": ");
                libs[i] = l.getLibrary(dll.getAbsoluteFile().getPath());
                System.out.println("OK");
            }

        } catch (LTDLException e) {
            e.printStackTrace();
            showWarning(
                "Problem loading library: "
                    + (dll != null ? "\n" + dll.toString() + "\n" : "")
                    + e.getMessage());
            loadError = true;
        } catch (Error e) {
            // it is not recommended to continue after catching an error
            // this catch statement displays a user message and exits
            e.printStackTrace();
            GUIUtils.userMessage(this, "Fatal Error", e.getMessage());
            System.exit(-1);
        }

        if (!loadError) {
            // display the recipe selector as there was no error loading the lib
            final JFrame frame = new JFrame();
            Cursor orig = getCursor();
            setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

            RecipeSelector rsel = new RecipeSelector(libs, !readOnly) {
                protected void approveSelection() {
                    Recipe recipe = getSelectedRecipe();
                    frame.dispose();

                    if (!readOnly) {
                        for (int i = 0; i < libs.length; i++) {
                            addLibrary(libs[i]);
                        }
                    }
                }
                protected void cancelSelection() {
                    frame.dispose();
                }
            };
            StringBuffer title = new StringBuffer();
            if (libs.length == 0) {
                title.append("No Libraries Selected");
            }
            if (libs.length == 1) {
                title.append("Contents of " + libs[0].getName());
            } else {
                title.append("Library Contents");
                /*
                for (int i = 0; i < libs.length; i++) {
                    if (i > 0) {
                        title.append(", ");
                    }
                    title.append(libs[i].getName());
                }
                */
            }

            // select the first recipe of the first library if possible.
            if (libs.length > 0
                && libs[0].getRecipes().length > 0
                && rsel.getTree().getRowCount() > 1) {
                rsel.getTree().setSelectionRow(1);
            }
            frame.setTitle(title.toString());
            frame.getContentPane().add(rsel);
            frame.pack();
            GUIUtils.center(this, frame);
            frame.addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent e) {
                    frame.dispose();
                }
            });
            frame.setVisible(true);

            setCursor(orig);
        }
        if (l != null) {
            try {
                releaseLoader();
            } catch (LTDLException e) {
                e.printStackTrace();
                showWarning("Problem unloading library: " + e.getMessage());
            }
        }

    }

    private String getDLLExtension() {
        String rc;
        String os = System.getProperty("os.name");
        if (os.equals("HP-UX")) {
            // HP-UX
            rc = ".sl";
        } else {
            // Linux and Solaris (SunOS).
            rc = ".so";
        }
        return rc;
    }

    /**
     * Delete a library from the table. Assumes that the table model is a ListAdaptor.
     */
    private void deleteSelectedLibrary(JTable table) {
        int selRow[] = table.getSelectedRows();

        if (selRow.length == 0) {
            return;
        }
        Arrays.sort(selRow);
        for (int i = selRow.length - 1; i >= 0; i--) {
            ((ListAdaptor) table.getModel()).removeRow(selRow[i]);
        }
        edited = true;
    }

    /**
     * Create a LibraryDescrption from the lib and add it to the
     * data List.
     * @param lib
     */
    private void addLibrary(RecipeLibrary lib) {
        List recipes = new ArrayList();
        for (int i = 0; i < lib.getRecipes().length; i++) {
            recipes.add(lib.getRecipes()[i].getName());
        }
        ((ListAdaptor) table.getModel()).appendRow(
            new LibraryDescription(lib.getLocation(), recipes));
        edited = true;
    }

    private void showWarning(String warning) {
        JOptionPane.showMessageDialog(
            this,
            warning,
            "Warning",
            JOptionPane.OK_OPTION);
    }

    private void showError(String msg) {
        JOptionPane.showMessageDialog(
            this,
            msg,
            "Error",
            JOptionPane.OK_OPTION);
    }

    /**
     * A class for containing the description of a library (path) and
     * content recipes.
     */
    static class LibraryDescription {
        private String libraryName;
        private List recipes; // List of String objects.

        /**
         * Create a Linrary Description.
         * @param name String name of the library (full path).
         * @param recs List of String recipe names
         */
        public LibraryDescription(String name, List recs) {
            libraryName = name;
            recipes = recs;
        }

        /**
         * Get the String used to display this linrary in the JTable.
         * @return String cell value.
         */
        public String toString() {
            String dispLibName = libraryName;
            // chop it to make it more displayable..
            if (dispLibName.length() > 60) {
                // take the first 25 and the last 25 characters
                dispLibName =
                    dispLibName.substring(0, 25)
                        + "......"
                        + dispLibName.substring(dispLibName.length() - 26);
            }
            StringBuffer rc = new StringBuffer(dispLibName + "      ");
            for (int x = 0; x < recipes.size(); x++) {
                if (x > 0) {
                    rc.append(", ");
                }
                rc.append(((String) recipes.get(x)).toUpperCase());
            }
            return rc.toString();
        }

        /**
         * Get the String used to store this information in GasProp.
         * i.e. rec1,rec2=LIB 
         * @return String
         */
        public String toPropertyString() {
            StringBuffer rc = new StringBuffer();
            for (int i = 0; i < recipes.size(); i++) {
                if (i > 0) {
                    rc.append(",");
                }
                rc.append((String) recipes.get(i));
            }
            rc.append("="+libraryName);
            return rc.toString();
        }
        /**
         * Returns the libraryName.
         * @return String
         */
        public String getLibraryName() {
            return libraryName;
        }

        /**
         * Returns the recipes.
         * @return List
         */
        public List getRecipes() {
            return recipes;
        }

    }
}
