package org.eso.ohs.phase2.apps.p2pp;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Dictionary;
import java.util.Hashtable;
import java.util.Stack;
import java.util.StringTokenizer;

import org.apache.log4j.Logger;
import org.eso.ohs.core.gui.widgets.ErrorMessages;
import org.eso.ohs.core.utilities.ConnectionPool;
import org.eso.ohs.core.utilities.FileIOUtils;
import org.eso.ohs.core.utilities.OHSKeys;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.BusinessObject;
import org.eso.ohs.dfs.CalibrationBlock;
import org.eso.ohs.dfs.ConstraintProperties;
import org.eso.ohs.dfs.FindingChart;
import org.eso.ohs.dfs.ObjectFactory;
import org.eso.ohs.dfs.ObservationBlock;
import org.eso.ohs.dfs.ObservationDescription;
import org.eso.ohs.dfs.ObservingRun;
import org.eso.ohs.dfs.StorableObject;
import org.eso.ohs.dfs.Summary;
import org.eso.ohs.instruments.Instrument;
import org.eso.ohs.instruments.InstrumentList;
import org.eso.ohs.instruments.TemplateVersionInfo;
import org.eso.ohs.instruments.ZipUtils;
import org.eso.ohs.p2pp.Folder;
import org.eso.ohs.persistence.AppConfig;
import org.eso.ohs.persistence.AutoSaver;
import org.eso.ohs.persistence.Config;
import org.eso.ohs.persistence.DirectoryEvent;
import org.eso.ohs.persistence.DirectoryNode;
import org.eso.ohs.persistence.Media;
import org.eso.ohs.persistence.OHSPersistence;
import org.eso.ohs.persistence.ObjectManager;
import org.eso.ohs.persistence.ObjectNotFoundException;
import org.eso.ohs.persistence.StorageManager;
import org.eso.ohs.persistence.StorageManager.IdType;
import org.eso.ohs.persistence.StorageManager.MultiObjSpec;

/**
 * This subclass of the ObjectManager handles all application-specific
 * issues relating to storage of objects.
 * 
 * <p>
 * For correct operation, there should be at most one instance of this
 * object.  Thus it is a singleton and can only be instantiated using
 * the static Persistence.getInstance() method.  Each time this
 * method is called it will return the <em>same</em> instance of the
 * ObjectManager.
 * 
 * <p>
 * <br>
 * @author Bob Kemp
 * <br>ChangeLog: cmavros 22Apr02 1195/P2PP Validation of OBs through P2PP AppServer.
 */
public class Persistence extends OHSPersistence {

    private static Logger stdlog_ = Logger.getLogger(Persistence.class);

    private final String rcsid =
        "$Id: Persistence.java,v 1.33 2005/11/02 11:22:08 tcanavan Exp $";

    private static Persistence instance_;

    private ObsRunRegistry obsRunRegistry_ = new ObsRunRegistry();

    /** current user id */
    protected int userId_ = -1;

    /**
     * The Dictionary of String names of Instruments with NO IP. These
     * instruments can not be down-loaded from the IP Archive and no attempt is
     * made to download them.
     */
    private Dictionary instrumentsWithNoIp = null;

    /**
     * Return the one and only instance of the ObjectManager.
     * 
     * @return Singleton ObjectManager instance.
     */
    public static Persistence getInstance() {
        return instance_;
    }

    /**
     * Install a new instance of the Persistence.
     * 
     * @param instance	new Persistence instance.
     */
    public static void setInstance(Persistence instance) {
        instance_ = instance;
        setObjectManager(instance);
    }

    public static void startup() {
        Persistence.setInstance(new Persistence());
        getInstance().startAutoSaver();
    }

    /**
     * Shutdown the current instance of ObjectManager and AutoSaver.
     */
    public void shutdown() {
        stopAutoSaver();
        ConnectionPool.shutdown();
        super.shutdown();
    }

    /**
     * AutoSaver creation hook -- overridden by unit tests
     * 
     * @see #startup
     * @see #shutdown
     */
    protected void startAutoSaver() {
        AutoSaver.startAutoSaver();
    }

    /**
     * AutoSaver shutdown hook -- overridden by unit tests
     * 
     * @see #startup
     * @see #shutdown
     */
    protected void stopAutoSaver() {
        AutoSaver.stopAutoSaver();
    }

    /**
     * Re-creation hook for subclasses.
     * 
     * @see #startup
     * @see OHSPersistence#restart
     * @see #shutdown
     */
    public void createNewInstance() {
        startup();
    }

    protected void dequeueObject(StorableObject sobj) {
        AutoSaver saver = AutoSaver.getAutoSaver();

        if (saver != null) {
            saver.removeFromQueue(sobj);
        }
    }

    public void queueObject(StorableObject sobj) {
        AutoSaver saver = AutoSaver.getAutoSaver();

        if (saver != null) {
            saver.addToQueue(sobj);
        }
    }

    /**
     * Add a link to this StorableObject into the Registry
     * so that it can be found quickly when needed.  This is used to
     * ensure that we only have one copy of an object in memory at
     * any one time!
     * 
     * @param device	medium on which the object is stored.
     * @param dir	directory in which the object is/should be stored.
     * @param bo	StorableObject with the id to be initialised.
     */
    protected void linkBusObj(
        Media device,
        DirectoryNode dir,
        boolean isNew,
        StorableObject bo) {
        super.linkBusObj(device, dir, isNew, bo);

        if (bo instanceof ObservingRun) {
            RegistryEntry re = registry_.lookup(bo.getId());
            obsRunRegistry_.addByName(bo.getName(), re);
        }
    }

    /**
     * Check a CalBlk/ObsBlk out of the dbase and update/create a local copy.
     * 
     * @param cb	Database CalibrationBlock or ObservationBlock
     *			to be checked-out.
     * 
     * @exception ObjectNotFoundException
     *			if the object was not found.
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */
    public void checkOut(CalibrationBlock cb)
        throws ObjectNotFoundException, ObjectIOException {
        StorageManager stMgr = getStorageManager(Media.DBASE);
        BusinessObject[] dbBOs = allDescendantsOf(cb);
        IdType[] idTypes = new IdType[dbBOs.length];
        CalibrationBlock locCB = null;
        FindingChart[] fcCharts;
        boolean createOB = false;
        FindingChart fcs[] = null;

        if (cb.getType().equals("O")) {
            fcs = populateFindingCharts(Media.DBASE, cb.getId());
            ((ObservationBlock) cb).setFindingCharts(fcs);
        }

        try {
            // If can read CalBlk/ObsBlk, then we have everything in memory.
            locCB =
                (CalibrationBlock) getBusObj(Media.LOCAL,
                    cb.getCloneParentId(),
                    cb.getClass());

            for (int j = 0; j < dbBOs.length; j++) {
                BusinessObject busObj = dbBOs[j];

                // Ensure local copy is uptodate !!!
                // Time-consuming but very important
                try {
                    refreshBusObj(
                        Media.PHASE2_DB,
                        busObj.getId(),
                        busObj.getClass());
                } catch (ObjectIOException e) {
                    ErrorMessages.announceIOError(null, e);
                    return;
                } catch (ObjectNotFoundException e) {
                    ErrorMessages.announceNoObject(null, e);
                    return;
                }

                // Copy dbase BO onto corresponding BO in local
                //  storage to capture any changes made to the
                //  dbase copy, eg by USG astronomers.
                BusinessObject locBusObj =
                    (BusinessObject) getBusObj(Media.LOCAL,
                        busObj.getCloneParentId(),
                        busObj.getClass());

                String[] propNames = busObj.simpleProperties();
                busObj.copyProperties(propNames, locBusObj);
            }
        } catch (InvocationTargetException e) {
            e.getTargetException().printStackTrace();
            throw new ObjectIOException(e.getTargetException());
        } catch (NoSuchMethodException e) {
            throw new ObjectIOException(e);
        } catch (IllegalAccessException e) {
            throw new ObjectIOException(e);
        } catch (ObjectNotFoundException e) {
            // no local copy - create it (after the checkout).
            stdlog_.debug("!!!!!!!!!!!!!!!Create OB");
            createOB = true;
            // Refresh the status of the database object just in 
            // case somebody has changed it in the OT
            refreshBusObj(
                Media.DBASE,
                cb.getId(),
                Config.getCfg().getClassFromId(cb.getId()));
        }
        for (int j = 0; j < dbBOs.length; j++) {
            idTypes[j] = new IdType(dbBOs[j].getId(), dbBOs[j].getClass());
            dequeueObject(dbBOs[j]);
        }

        try {
            stMgr.checkOut(idTypes);
        } catch (ObjectIOException e) {
            // remove from memory as there was an error with the object.
            for (int j = 0; j < idTypes.length; j++) {
                registry_.remove(idTypes[j].id);
            }
            throw e;
        } catch (ObjectNotFoundException e) {
            // remove from memory as there was an error with the object.
            for (int j = 0; j < idTypes.length; j++) {
                registry_.remove(idTypes[j].id);
            }
            throw e;
        }

        for (int j = 0; j < idTypes.length; j++) {
            registry_.remove(idTypes[j].id);
        }
        stdlog_.debug("Should object be removed");
        if (cb.getStatus().equals("D") || cb.getStatus().equals("-")) {
            // Show user that OB is checked-in.
            stdlog_.debug("Object is Defined ");

            if (createOB) {
                stdlog_.debug("!!!!!!!!!Creating OB");
                try {

                    ObservingRun or = cb.getObsRun();
                    locCB =
                        (CalibrationBlock) deepCopy(Media.LOCAL,
                            getDirectoryFor(or.getId()),
                            cb,
                            cb.getCloneParentId());
                    locCB.setDbaseId(cb.getId());
                    if (locCB instanceof ObservationBlock) {
                        if (fcs.length > 0) {
                            ObservationBlock ob = (ObservationBlock) locCB;
                            String folderName =
                                getRootFolderPath(locCB.getObsRun());
                            for (int i = 0; i < fcs.length; i++) {
                                File f =
                                    new File(
                                        folderName
                                            + File.separator
                                            + fcs[i].getCheckSum()
                                            + "."
                                            + fcs[i].getFcType());
                                if (!f.exists()) {
                                    FileIOUtils.writeBinaryFile(
                                        fcs[i].getFcData(),
                                        f);
                                }
                            }
                        }
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
            locCB.setStatus(CalibrationBlock.PARTIALLY_DEFINED);
            locCB.recursiveMarkCheckedOut();
            if (locCB.getType().equals("O")) {
                ((ObservationBlock) locCB).setFindingCharts(fcs);
                for (int i = 0; i < fcs.length; i++) {
                    fcs[i].setFcData(null);
                }
            }

        }
    }

    public boolean isInstrumentWithNoIP(String instrument) {
        if (instrumentsWithNoIp == null) {
            instrumentsWithNoIp = getInstrumentsWithNOIP();
        }
        return instrumentsWithNoIp.get(instrument) != null;
    }

    /**
     * Download and extract the specified IP from the database into the IPCache directory,
     * and add it to the InstrumentList.
     * There is no check if the IP is aleary in the cache.
     * @throws IOException If the directory was not configured, the database coneection and user info is not
     * configured or there is a database or IO failure.
     * @return boolean true if downloaded, else false (ignored instrument or not found).
     */
    public boolean downloadInstrumentPackage(String instrument, float version)
        throws ObjectIOException {

        if (isInstrumentWithNoIP(instrument)) {
            // this instrument has no ip - so no download is attempted.
            return false;
        }
        StorageManager stMgr = getStorageManager(Media.DBASE);
        String ipCacheDir = AppConfig.getAppConfig().getIPVersionIPCache();
        if (ipCacheDir == null || ipCacheDir.length() == 0) {
            throw new ObjectIOException(
                "Instrument Cache directory is not "
                    + " configured in the config files. "
                    + " config key: "
                    + OHSKeys.IPVersioningIPCache);
        }
        File f = new File(ipCacheDir);
        ByteArrayInputStream is =
            new ByteArrayInputStream(
                stMgr.downloadInstrumentPackage(instrument, version));
        try {
            ZipUtils.extract(is, f.getAbsolutePath(), null, true);
        } catch (IOException e) {
            throw new ObjectIOException(e);
        }
        return true;
    }

    /**
     * @return Vector of String names of instruments that are 
     * not downloadable as they do not have IPs (known not to have IPs
     * in the IP archive. This list is taken from the configuration data (site.
     * cf) An empty Vector is returned if no ignored instruments are found.
     */

    public Dictionary getInstrumentsWithNOIP() {
        Dictionary rc = new Hashtable();
        String list =
            AppConfig.getAppConfig().getValue(OHSKeys.InstrumentsWithNoIP);
        if (list != null && list.length() > 0) {
            String delim = " ,;";
            StringTokenizer st = new StringTokenizer(list, delim);
            while (st.hasMoreTokens()) {
                String inst = st.nextToken();
                rc.put(inst, inst);
            }
        }
        return rc;
    }

    /**
     * Validation of OBs through P2PP AppServer.
     * 
     * @param cb	CalibrationBlock or ObservationBlock to be checked-in.
     * 
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     * <br>ChangeLog: cmavros 22-Apr-2002
     */
    public boolean verifyObject(CalibrationBlock cb) throws ObjectIOException {

        DirectoryNode dbRoot = getRoot(Media.DBASE);
        CalibrationBlock dbCB =
            (CalibrationBlock) deepCopy(Media.DBASE,
                dbRoot,
                cb,
                cb.getDbaseId());

        try {
            MultiObjSpec[] mos = toMultiObjSpecArray(dbCB);
            RegistryEntry oi = registry_.lookup(dbCB.getId());
            StorageManager stMgr = oi.getStorageManager();
            return (stMgr.verifyObject(mos));
        } catch (ObjectIOException e) {
            throw e;
        } finally {
            forgetBusObj(Media.DBASE, dbCB.getId(), dbCB.getClass());
        }

    }

    private void setFcData(ObservationBlock ob,Hashtable fcCrcList)
        throws ObjectNotFoundException, IOException {
        FindingChart[] fcCharts = ob.getFindingCharts();
        for (int i = 0; i < fcCharts.length; i++) {
            long checkSum = fcCharts[i].getCheckSum();
            ObservingRun obsRun = ob.getObsRun();
            String path = getRootFolderPath(obsRun);
            if (!fcCrcList.containsKey(new Long(checkSum))) {
                fcCharts[i].setFcData(fcCharts[i].getFcDataLocal(path));
            } else {
                fcCharts[i].setFcData(null);             
            }
        }
    }

    /**
     * Check a CalBlk/ObsBlk and it's associated BOs, etc into the database.
     * 
     * @param cb	CalibrationBlock or ObservationBlock to be checked-in.
     * 
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     * <br>ChangeLog: Toni Schlichting 22-June-2001
     */
    public void checkIn(CalibrationBlock cb,Hashtable fcCrcList)
        throws ObjectIOException, ObjectNotFoundException, IOException {

        DirectoryNode dbRoot = getRoot(Media.DBASE);
        CalibrationBlock dbCB =
            (CalibrationBlock) deepCopy(Media.DBASE,
                dbRoot,
                cb,
                cb.getDbaseId());

        try {
            if (dbCB.getType().equals("O")) {
                setFcData((ObservationBlock) dbCB,fcCrcList);
            }
            // Checked-in OBs take "D" status.
            dbCB.setStatus(CalibrationBlock.DEFINED);

            MultiObjSpec[] mos = toMultiObjSpecArray(dbCB);
            RegistryEntry oi = registry_.lookup(dbCB.getId());
            StorageManager stMgr = oi.getStorageManager();
            stMgr.checkIn(mos);

            for (int j = mos.length - 1; j >= 0; j--) {
                StorableObject sobj = mos[j].obj.getObject();
                RegistryEntry re = registry_.lookup(sobj.getId());
                re.setNew(false);
            }
            setDbaseIds(cb, dbCB);

            // Show user that OB is checked-in.
            cb.setStatus(CalibrationBlock.DEFINED);
            cb.setStatus(CalibrationBlock.DEFINED);

            cb.recursiveMarkCheckedIn();
            if (dbCB.getType().equals("O")) {
                FindingChart[] fcCharts =
                    ((ObservationBlock) cb).getFindingCharts();
                for (int i = 0; i < fcCharts.length; i++) {
                    fcCharts[i].setFcData(null);
                }
            }
        } catch (ObjectIOException e) {
            throw e;
        } finally {
            forgetBusObj(Media.DBASE, dbCB.getId(), dbCB.getClass());
        }

    }
    
    /**
     * Check in a readme file 
     * 
     * @param cb    CalibrationBlock or ObservationBlock to be checked-in.
     * 
     * @exception ObjectIOException
     *          if there was an I/O error writing the object.
     * <br>ChangeLog: Toni Schlichting 22-June-2001
     */
    public void usdSubmit(ObservingRun obsRun)
        throws ObjectIOException, ObjectNotFoundException, IOException {
        try {
            StorageManager stMgr = getStorageManager(Media.DBASE);
            stMgr.usdSubmit(obsRun);
        } catch (ObjectIOException e) {
            throw e;
        } 
    }

    
    
    
    /**
     * Check a CalBlk/ObsBlk and it's associated BOs, etc into the database.
     * 
     * @param cb    CalibrationBlock or ObservationBlock to be checked-in.
     * 
     * @exception ObjectIOException
     *          if there was an I/O error writing the object.
     * <br>ChangeLog: Toni Schlichting 22-June-2001
     */
    public String checkOutReadmeFile(ObservingRun obsRun)
        throws ObjectIOException, ObjectNotFoundException, IOException {
        try {
            StorageManager stMgr = getStorageManager(Media.DBASE);
            String xmlfile = stMgr.checkOutReadme(obsRun);
            return xmlfile;
        } catch (ObjectIOException e) {
            throw e;
        } 
    }


    

    /**
     * Verify that the TemplateSignature files used by the StorageManager
     * are the same version as those specified (and used by the client)
     * 
     * @return	list of the names of TemplateSignatures that are out of date.
     * 
     * @exception ObjectIOException
     *			if there was an I/O error while checking
     */
    public String[] verifyTemplateSignatures(
        Media device,
        TemplateVersionInfo[] tsList)
        throws ObjectIOException {

        return getStorageManager(device).verifyTemplateSignatures(tsList);
    }

    /**
     * Returns the directory in which the object with the given id is stored.
     * Throws an IllegalArgumentException if it is not currently in the
     * Registry.
     *
     * @param id	id of StorableObject whose containing directory
     * 			is required
     * 
     * @return		Directory containing the object
     * 
     * @exception IllegalArgumentException
     * 			if the object is not known to the system,
     * 			ie not in Registry.
     */
    public DirectoryNode getDirectoryFor(long id)
        throws IllegalArgumentException {
        RegistryEntry oi = registry_.lookup(id);

        if (oi == null) {
            throw new IllegalArgumentException("Object not in Registry: " + id);
        }

        return oi.getDirectory();
    }

    /**
     * Returns the directory in which the object is stored.  Throws an
     * IllegalArgumentException if it is not currently in the Registry.
     *
     * @param sobj	StorableObject whose containing directory is required
     * 
     * @return		Directory containing the object
     * 
     * @exception IllegalArgumentException
     * 			if the object is not known to the system,
     * 			ie not in Registry.
     */
    public DirectoryNode getDirectoryFor(StorableObject sobj)
        throws IllegalArgumentException {
        RegistryEntry oi = registry_.lookup(sobj.getId());

        if (oi == null) {
            throw new IllegalArgumentException(
                "Object not in Registry: " + sobj);
        }

        return oi.getDirectory();
    }

    /**
     * Returns the named ObservingRun.  Throws an IllegalArgumentException
     * if it is not currently in the Registry.
     *
     * @param name	Name of an observing run
     * 
     * @return		Corresponding ObservingRun object
     * 
     * @exception IllegalArgumentException
     * 			if the object is not known to the system,
     * 			ie not in Registry.
     */
    public ObservingRun getObservingRunByName(String name)
        throws IllegalArgumentException {
        RegistryEntry oi = obsRunRegistry_.lookupByName(name);

        if (oi == null) {
            throw new IllegalArgumentException(
                "Observing run not in Registry: " + name);
        }

        return (ObservingRun) oi.getObject();
    }

    /**
     * Test for whether the user is allowed to delete the directory.
     * Used by deleteDirectory().
     * 
     * @param device A reference to the selected media. 
     * @param node The directory to be deleted on the selected media
     *
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */
    protected boolean canDeleteDirectory(Media device, DirectoryNode node)
        throws ObjectIOException {
        if (!super.canDeleteDirectory(device, node)) {
            return false;
        }

        Summary[] summ = listObjects(device, node);

        if (summ.length > 1
            || (summ.length == 1 && summ[0].getObjectClass() != Folder.class)) {
            return false;
        }

        return true;
    }

    /**
     * Hook to allow an application-specific subclass to
     * prepare a directory for deletion, eg cleaning out
     * application-specific data contained in the directory.
     * 
     * @param device A reference to the selected media. 
     * @param node The directory to be deleted on the selected media
     *
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */
    protected void prepareDirectoryForDeletion(
        Media device,
        DirectoryNode node)
        throws ObjectIOException {
        RegistryEntry oi = registry_.lookup(node.getId());

        if (oi != null) {
            try {
                deleteBusObj(device, node.getId(), Folder.class);
            } catch (ObjectNotFoundException e) {
                stdlog_.error(e);
                // We don't care if the summary file wasn't found.
            }
        }
    }

    /**
     * Creates a directory on external media.
     *
     * @param device A reference to the selected media. 
     * @param node The directory to be created on the selected media 
     * @return True if directory created successfully otherwise false      
     *
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */
    public long createDirectory(Media device, DirectoryNode node)
        throws ObjectIOException {
        long id = getNewId(device, Folder.class);

        return createDirectory(device, node, id);
    }

    /**
     * Renames a directory on external media.
     *
     * @param device	reference to the selected media. 
     * @param node	directory to be created on the selected media.
     * @param newName	new (user-level) name for the directory.
     */
    public void renameDirectory(
        Media device,
        DirectoryNode node,
        String newName) {
        node.setDescription(newName);
        fireDirectoryEvent(
            DirectoryEvent.DIRECTORY_CHANGED,
            device,
            node,
            null);
    }

    /**
     * Check in a readme file 
     * 
     * @param cb    CalibrationBlock or ObservationBlock to be checked-in.
     * 
     * @exception ObjectIOException
     *          if there was an I/O error writing the object.
     * <br>ChangeLog: Toni Schlichting 22-June-2001
     */
    public void checkInReadmeFile(ObservingRun obsRun) throws ObjectIOException, ObjectNotFoundException, IOException {
        try {
            StorageManager stMgr = getStorageManager(Media.DBASE);
            stMgr.checkInReadme(obsRun);
        } catch (ObjectIOException e) {
            throw e;
        } 
    }

    /**
     * A Registry-like table for cached ObservingRun instances.
     * ObservingRuns are referred to here by name, not id.
     */
    private static class ObsRunRegistry {
        /** Hashtable storing the ObservingRun objects */
        private final Hashtable byName_ = new Hashtable();

        /**
         * Add the specified RegistryEntry to the observing run registry.
         * Any previous instance is replaced.
         * 
         * @param name	name of the ObservingRun, eg "63.H-0420(A)"
         * @param oi	RegistryEntry object describing this ObservingRun
         */
        public void addByName(String name, RegistryEntry oi) {
            byName_.put(name, oi);
        }

        /**
         * Returns the RegistryEntry corresponding to the observing run
         * with the given name.
         * 
         * @param name	name of the ObservingRun, eg "63.H-0420(A)"
         * @return	RegistryEntry object describing this ObservingRun,
         * 		or null if not found
         */
        public RegistryEntry lookupByName(String name) {
            return (RegistryEntry) byName_.get(name);
        }

    }
        
    /**
     * Is this object editable (from a client standpoint) i.e.:
     * checks its status - if it is checked in, it is not editable.
     */

    public boolean isAllowedToModify(BusinessObject bo) {
        if (bo == null || !isInRegistry(bo.getId()))
            return false;
        Media media = Config.getCfg().getMediaFromId(bo.getId());
        return !bo.isCheckedIn() && Media.LOCAL.equals(media);

    }
    
    /**
     * Creates a StorableObject, registers it with the ObjectManager
     * and, if it's a CalBlk or ObsBlk, sets the ObservingRun property.
     * 
     * @param device    Media on which the object is to be stored
     * @param dir   Directory in which the object is to be stored
     * @param objType   Type of object to be created
     * 
     * @exception ObjectIOException Object I/O Excpetion     
     */
    public BusinessObject createAndRegister(
        Media device,
        DirectoryNode dir,
        Class objType)
        throws ObjectIOException {

        BusinessObject sobj = ObjectFactory.create(objType);

        if (device.equals(Media.LOCAL))
            initBusObj(device, dir, sobj);
        register(device, dir, sobj);

        return sobj;
    }

    public void initBusObj(
        Media device,
        DirectoryNode node,
        BusinessObject sobj)
        throws ObjectIOException {
        stdlog_.info("p2pp.MyPersistCfg: Init BUS OBJ called");
        Folder folder = FolderFactory.getFolder(device, node);
        if (sobj instanceof CalibrationBlock) {
            CalibrationBlock cb = (CalibrationBlock) sobj;

            cb.setObsRun(folder.findObsRun());
        } else if (sobj instanceof ObservationDescription) {
            ObservationDescription od = (ObservationDescription) sobj;

            // set the OD instrument and version from the InstrumentList,
            // not from what the folder says. In Single IP mode, another IP
            // may be returned and the OD should contain the actual IP version
            // used.

            Instrument inst =
                InstrumentList.getInstance().getInstrument(
                    folder.findObsRun().getInstCode(),
                    folder.findObsRun().getIPVersion());
            od.setInstrument(inst.getInstrumentName(), inst.getVersionNumber());
        }

    }

    /** 
     * Create and Register business objects and all its children
     * @param device Media to save data
     * @param dir Directory to save data in
     * @param objType ObjType to be saved
     * @return Business object
     */

    public BusinessObject createAndRegisterRecursive(
        Media device,
        DirectoryNode dir,
        Class objType)
        throws ObjectIOException {
        BusinessObject bo = ObjectFactory.createRecursive(objType);
        Stack stk = new Stack();

        stk.push(bo);
        while (!stk.isEmpty()) {
            BusinessObject busObj = (BusinessObject) stk.pop();

            BusinessObject[] kids = busObj.getBusObjChildren();
            for (int j = 0; j < kids.length; j++) {
                stk.push(kids[j]);
            }
            if (device.equals(Media.LOCAL))
                initBusObj(device, dir, busObj);
        }
        ObjectManager.getObjectManager().recursiveRegister(device, dir, bo);
        if (bo instanceof ObservationBlock) {
            ObservationBlock ob = (ObservationBlock) bo;
            try {
				ConstraintProperties.setConstraintDefaults(ob);
			} catch (NoSuchMethodException e) {
				stdlog_.debug(e.getMessage());
			} catch (IllegalAccessException e) {
				stdlog_.debug(e.getMessage());
			} catch (InvocationTargetException e) {
				stdlog_.debug(e.getMessage());
			}
        }
        return bo;
    }
}
